var express            = require('express');
var app                = express();
var session            = require('express-session');
var https              = require('https');
var http               = require('http');
var fs                 = require('fs');
var path               = require('path');
var co                 = require('co');
var colors             = require('colors');
var syntaxError        = require('syntax-error');
var assert             = require('assert');
var __pointerStore     = require('./imports' ).__pointerStore;
var getCallerFile      = require('./imports' ).getCallerFile;
var Database           = require('./database');
global.imports         = require('./imports' ).imports;
global.namespace       = require('./imports' ).namespace;
global.Intuition       = require('./intuition.js');
global.Router          = require('./router.js' );


/**
 * @class Builder - is responsible for handling the application components and linking the classes of the project to
 * the express engine.
 * */
var Builder = function() {
  /**
   * the compiledNamespacesHash is responsible for keeping track of files that have already been jit compiled,
   * (i.e. required was successfully run on that file)
   *
   *  @param #compiledNamespacesHash#filename
   *  @type {String}
   *  @default undefined
   *
    @name Builder#compiledNamespacesHash
    @type {Object}
    @default {}

   * the applicationList is an array that keeps track of all the ApplicationObjects used in the Intuition project. An
   * ApplicationObject is a model used to describe a project dependency
   * @class ApplicationObject
   *
   *
   * the ApplicationObject#parentString describes the string passed in for the namespace of the Object. For examples
   * Some.Name.Space
   * @param ApplicationObject#parentString
   * @type {String}
   * @default undefined
   *
   * the ApplicationObject#parentPath describes the split by '.' of the parentPath, e.g. ['Some', 'Name', 'Space']
   * @param ApplicationObject#parentPath
   * @type {Array<String>}
   * @default []
   *
   * the ApplicationObject#namespaceString describes the string passed in for the namespace of the Object and the
   * Object Name. For example Some.Name.Space.MyObject
   * @param ApplicationObject#namespaceString
   * @type {String}
   * @default undefined
   *
   * the ApplicationObject#hashPath describes the split by '.' of the namespaceString, for example
   * ['Some', 'Name', 'Space', 'MyObject']
   * @param ApplicationObject#parentPath
   * @type {Array<String>}
   * @default []
   *
   *
    @name Builder#applicationList
    @type {Array<ApplicationObjects>}
    @default []
   *
   * the dependencyQueue is responsible for keeping track of ApplicationObjects that could not be built because they
   * require a dependency that has not been compiled yet
   *
    @name Builder#dependencyQueue
    @type {Array<ApplicationObjects>}
    @default []
   *
   * the objectMap is the list of filenames pointing at their compiled Intuition Classes. This is the brain of the 
   * framework
   *
    @name Builder#objectMap
    @type {Object<String>{IntuitionClass}}
    @default {}
   *
   *
   * the lastRequiredApplicationObject is the object last required by an Intuition Class. This is required when building
   * the dependency queue
   *
    @name Builder#lastRequiredApplicationObject
    @type {ApplicationObject}
    @default null
   *
   * intuitionDirectories are the directories the Builder should traverse to find the intuition objects included in the
   * project
   *
    @name Builder#intuitionDirectories
    @type {Array<String>}
    @default ['models', 'services', 'controllers', 'views']
   *
   * templateDirectory is the directory the builder traverses to find all the template views required in the project
   *
    @name Builder#templateDirectory
    @type {String<String>}
    @default ['models', 'services', 'controllers', 'views']
   *
   * templateHash is a list of all the template contents by their filenames relative to the templateDirectory
   *
    @name Builder#templateHash
    @type {Object{String}{template}}
    @default {}
   *
   *
  */
  this.compiledNamespacesHash = {};
  this.applicationList = [];
  this.dependencyQueue = [];
  this.compiledObjectDirectory  = {};
  this.objectMap = {};
  this.lastRequiredApplicationObject = null;
  this.intuitionDirectories = ['models', 'services', 'controllers', 'views'];
  this.templateDirectory = 'views/templates';
  this.templateHash = {};
};

Builder.prototype.buildApplicationList = function() {
  this.travseFileSystem();
  this.tryToOrderApplicationFromCache();
  this.compileNamespaces();
  this.traverseTemplates();
  var View = this.getObjectAtPath('Views.View'.split('.'));
  View.templates = this.templateHash;
}

Builder.prototype.traverseTests = function() {
  return FileSystemTraverser.getFileHashList('tests');
}

/**
 * traverseTemplates traverses all the template files with .html extension in the templateDirectory
 * Builder#traverseTemplates
 * @type {Function}
 * @return void
 * */

Builder.prototype.travseFileSystem = function() {

  for(var i = 0; i < this.intuitionDirectories.length; ++i) {
    var dir = this.intuitionDirectories[i];
    this.intuitionDirectories[i] = cwd + '/' + dir;
  }
  //include the files in the node module
  this.intuitionDirectories.push(__dirname + '/intuition');


  for (var i = 0; i < this.intuitionDirectories.length; ++i) {
    var files = Builder.FileSystemTraverser.getFileHashList(this.intuitionDirectories[i]);
    this.applicationList = this.applicationList.concat(files);
  }
}

Builder.prototype.tryToOrderApplicationFromCache = function() {
  try {
    var compiledApplicationFile = fs.readFileSync( cwd + '/.intuition/application.json', 'utf-8');
    var compiledApplication = JSON.parse(compiledApplicationFile);
    this.reorderApplication(compiledApplication);
  } catch (e) {
    console.log(e);
  }
}

/**
 * traverseTemplates traverses all the template files with .html extension in the templateDirectory
 * Builder#traverseTemplates
 * @type {Function}
 * @return void
 * */
Builder.prototype.traverseTemplates = function() {

  var templates = new Builder.FileSystemTraverser(this.templateDirectory, 'html');
  for (var i = 0; i < templates.files.length; ++i) {
    var key = templates.files[i].filename.replace(this.templateDirectory + '/', '');
    this.templateHash[key] = fs.readFileSync(templates.files[i].filename, 'utf-8');
  }
}

Builder.prototype.compileNamespaces = function() {

  for (var i = 0; i < this.applicationList.length; ++i) {
    var filename = this.applicationList[i].filename;
    if ( !this.compiledNamespacesHash[filename]) {
      try {
        this.compiledObjectDirectory[filename] =  require(filename);
        if (filename === this.applicationList[i].filename &&
                (!this.compiledObjectDirectory[filename] || !this.compiledObjectDirectory[filename].extend)
                ) {
          var src = fs.readFileSync(filename);
          var exportsPattern = new RegExp("\\s+module\\s*\\.\\s*exports\\s*=\\s*" + this.applicationList[i].object + "\\s*;?");
          if (!exportsPattern.test(src)) {
            console.log("Please add \"module.exports = " + this.applicationList[i].object.red + ";\" to %s".red, filename);
            process.exit();
          }
        }
        this.compiledObjectDirectory[filename].__name__ = this.applicationList[i].object;

      } catch (e) {
        if ( e instanceof SyntaxError) {
          var src = fs.readFileSync(filename);
          var err = syntaxError(src, filename);
          console.error('SYNTAX ERROR IN COMPILATION' + Array(62).join('!'));
          console.error(err);
          console.error(Array(76).join('-'));
          process.exit();
        }
        this.addToDependencyQueue(this.applicationList[i]);
        return
      }

      if (!this.failedCompilation) {
        this.setObjectAtPath(this.applicationList[i].hashPath, this.compiledObjectDirectory[filename]);
      }
    }
  }

  if(this.failedCompilation) {

    this.emptyDependencyQueue();
    this.compiledObjectDirectory = {};
    this.compiledNamespacesHash = {};
    this.failedCompilation = false;
    this.compileNamespaces();
    return;
  }

  for (let i = 0; i < this.applicationList.length; ++i) {
    (function(application){
      this.setObjectAtPath(this.applicationList[i].hashPath, null);

      this.createGlobalFeauturesForApplicationObject(this.applicationList[i]);
      this.createPlaceHolderObject(this.applicationList[i]);
    }).call(this, this.applicationList[i]);
  }
};

Builder.prototype.addToDependencyQueue = function(dependentObject) {
  var independentObject = this.lastRequiredApplicationObject;
  this.failedCompilation = true;

  if (!this.dependencyQueueContains(independentObject)) {
    if (!this.dependencyQueueContains(dependentObject)) {
      this.dependencyQueue = this.dependencyQueue.concat([independentObject, dependentObject]);
    } else {
      for (var i = 0; i < this.dependencyQueue.length; ++i) {
        if (this.dependencyQueue[i].filename == dependentObject.filename) {
          this.dependencyQueue.splice(i, 0, independentObject);
          break;
        }
      }
    }
  } else {
    //we are ignoring the double dependency case
    for (var i = 0; i < this.dependencyQueue.length; ++i) {
      if (this.dependencyQueue[i].filename == independentObject.filename) {
        this.dependencyQueue.splice(i + 1, 0, dependentObject);
        break;
      }
    }
  }
}

Builder.prototype.dependencyQueueContains = function(dependency) {
  for (var i = 0; i < this.dependencyQueue.length; ++i) {
    try {
      if(dependency.filename === this.dependencyQueue[i].filename) return true;
    } catch (e) {
      console.error(this.dependencyQueue);
      console.error(dependency)
      console.error("index " + i);
      console.error(e);
      console.error(getTrace(e));
      process.exit();
    }

  }
  return false;
};

Builder.prototype.emptyDependencyQueue = function() {
  for (var i = 0; i < this.dependencyQueue.length; ++i) {
    for(var j = 0; j < this.applicationList.length; ++j) {
      if (this.dependencyQueue[i].filename === this.applicationList[j].filename) {
        this.applicationList.splice(j,1);
        break;
      }
    }
  } //end outer for
  this.applicationList = this.dependencyQueue.concat(this.applicationList);
  this.dependencyQueue = [];
}

Builder.prototype.reorderApplication = function(compiledApplication) {

  // index the objects that have already been compiled before
  compiledApplication.forEach(function(asset, index) {
      return asset.$index = index;
  });

  /**
   * asset is an asset in the applicationList. If it's already been compiled before getAssetFromCompiledApplication
   * will return the compiled object, which will have an index associated with it
   * */
  var getAssetFromCompiledApplication = function(asset) {
    for(var i = 0; i < compiledApplication.length; ++i) {
      if(asset.filename === compiledApplication[i].filename) {
        return compiledApplication[i];
      }
    }
    return null;
  }

  this.applicationList.sort(function(compiledAssetA, compiledAssetB) {
     var a = getAssetFromCompiledApplication(compiledAssetA);
     var b = getAssetFromCompiledApplication(compiledAssetB);

     if(a !== null && b !== null) {
       return a.$index > b.$index ? 1 : -1;
     } else if (a && b === null) {
       return -1;
     } else if (a === null && b) {
       return 1;
     } else {
       return 0;
     }
  });

}


/**
 * this is called from the namespace object
 * */
Builder.prototype.compileNamespaceObject = function(namespaceObject) {
  for (var i = 0; i < this.applicationList.length; ++i) {
    var filename = this.applicationList[i].filename;
    if (filename === namespaceObject._callerfile) {
        if (!this.compiledNamespacesHash[filename]) {
          //add filenames to this.compiledNamespacesHash before calling compileNamespaces again
          this.compiledNamespacesHash[filename] = true;
          this.applicationList[i].parentString = namespaceObject._namespace;
          this.applicationList[i].parentPath = namespaceObject._namespace.split('.');
          this.applicationList[i].hashPath = this.applicationList[i].parentPath.concat(this.applicationList[i].object);
          this.applicationList[i].namespaceString = this.applicationList[i].hashPath.join('.');
          this.compileNamespaces();
        }
      break;
    }
  }
}

Builder.prototype.compileGlobalFeatures = function(namespaceObject) {
  for (var i = 0; i < this.applicationList.length; ++i) {
    var filename = this.applicationList[i].filename;
    if (filename === namespaceObject._callerfile) {
      this.createGlobalFeauturesForApplicationObject(this.applicationList[i]);
      break;
    }
  }
}

Builder.prototype.createGlobalFeauturesForApplicationObject = function(application) {
   var parentDirectory = this.getObjectAtPath(application.parentPath);
   for (var object in parentDirectory) {

     (function(object, application, parentDirectory){

        if(!__pointerStore[object]) {
          __pointerStore[object] = {};
        }
        __pointerStore[object][application.filename] = this.getObjectAtPath(application.parentPath);

       var lastCaller = null

       if(!global[object]) {
         Object.defineProperty(global, object , {
            get: function() {
              var callerfile = getCallerFile();
              if(callerfile && callerfile !== __filename && __pointerStore[object][callerfile]) {
                return __pointerStore[object][callerfile][object];
              }
              return null;
            },
            configurable: true
         });
       }

     }).call(this, object, application, parentDirectory);

   }
}

Builder.prototype.createPlaceHolderObject = function(application) {
  var self = this;
  var node = this.objectMap;
  var hashPath = application.parentPath;
  for (var i = 0; i < hashPath.length; ++i) {
     if(!node[hashPath[i]]) node[hashPath[i]] = {};
     node = node[hashPath[i]];
  }


  Object.defineProperty(node, application.object, {
    get: function() {
      self.lastRequiredApplicationObject = application;
      return self.compiledObjectDirectory[application.filename] ;

    },
    set: function(value) {
      self.compiledObjectDirectory[application.filename]  = value;
    },
    configurable: true,
  });

}

Builder.prototype.setObjectAtPath = function(path, object) {
  var node = this.objectMap;
  var hashPath = path;
  for (var i = 0; i < hashPath.length; ++i) {
     if (!node[hashPath[i]]) node[hashPath[i]] = {};
     if (i === i.length -1) {
       node[hashPath[i]] = object;
     } else {
       node = node[hashPath[i]];
     }
  }
}

Builder.prototype.getObjectAtPath = function(path) {
  var node = this.objectMap;
  var hashPath = path;
  for (var i = 0; i < hashPath.length; ++i) {
    node = node[hashPath[i]];
  }
  return node;
}

Builder.prototype.startServerWithDatabase = function(port, host, db) {
  this.db = db.database;
  this.store = db.store;
  global._database = db;
  this.startServer(port, host);
}
Builder.prototype.startServer = function(port, host) {

  var databaseJsonPath   = cwd + '/config/database.json';
  config             = require(databaseJsonPath);

  if (!port) port = '8000';
  if (!host) host = '127.0.0.1';

  // Catch errors
  this.store.on('error', function(error) {
    assert.ifError(error);
    assert.ok(false);
  });

  app.use(session({
    secret: config['session-secret'],
    cookie: {
      maxAge: config['session-expiration']
    },
    name: '_session_id',
    store: this.store,
    proxy: true,
    resave: true,
    saveUninitialized: true
  }));

  var routesArray = global.Router.routesToArray();
  var self = this;

  for(var i = 0; i < routesArray.length; ++i) {
    for(var http in routesArray[i].http) {
      (function(route, http){
        var urlArgs  = [];
        var urlString = route.split('\/');

        urlString.forEach(function(urlComponent) {
            if (urlComponent.substr(0,1) === ':') {
              urlArgs.push(new String);
            }
        });

        app.param(urlArgs, new Function);

        app[http.toLowerCase()](route, function(request, response, next) {
           var _routeInfo = global.Router.routes[route][http];
           var Controller = self.getObjectAtPath(_routeInfo['controller'].split('.'));
           var Response = self.getObjectAtPath('Intuition.Response'.split('.'));
           var Request = self.getObjectAtPath('Intuition.Request'.split('.'));
           var intuitionRequest = Request.new(request);
           var intuitionResponse = Response.new(response, intuitionRequest);
           var instance = Controller.new(intuitionRequest, intuitionResponse);
           var urlString = route.split('\/');
           var urlArgs  = [];

           urlString.forEach(function(urlComponent) {
             if(urlComponent.substr(0,1) === ':') {
               urlArgs.push(request.params[urlComponent.substr(1)]);
             }
           });

          var generatorMethod = instance[_routeInfo['method']];
          var argv = [instance].concat(urlArgs);
          var generator = generatorMethod.bind.apply(generatorMethod, argv);

          co( generator )
          .catch(function(error) {
             var Response = self.getObjectAtPath('Intuition.Response'.split('.'));
             var Request = self.getObjectAtPath('Intuition.Request'.split('.'));
             var intuitionRequest = Request.new(request);
             var intuitionResponse = Response.new(response, intuitionRequest);
             var ErrorController = self.getObjectAtPath('Controllers.IntuitionErrorController'.split('.'));
             var errorController = ErrorController.new(intuitionRequest, intuitionResponse);
             var generatorMethod = errorController.error;
             var generator = generatorMethod.bind.apply(generatorMethod, [errorController, error, 500]);
             console.error(error);
             co( generator )
          });
      });
      })(routesArray[i].route, http)

    } //end for http
  }//end for route

  app.use('/static', express.static(cwd + '/public'));

  app.listen(port, host, function () {
    console.log('listening on %s:%s!'.green, host, port);
  });
};

Builder.prototype.startDatabase = function() {
   this.dbInitializer = new Database();
   return this.dbInitializer.connection;
}

Builder.build = function() {
  var t1 = (new Date()).getTime();
  var _builder = new Builder();
  global.namespace.builder = _builder;
  imports.builder = _builder;
  _builder.buildApplicationList();
  fs.writeFileSync(cwd + '/.intuition/application.json', JSON.stringify(_builder.applicationList), 'utf-8');
  var t2 = (new Date()).getTime();
  console.log("compilation time", colors.blue( (t2 -t1).toString()));
  return _builder;
}

/**
 * FileSystemTraverser recursively traverses all files in given path
 * @class FileSystemTraverser
 * @param {String} path - the filepath to traverse
 * @param {String} extension - the fileextension to traverse for
 * */
var FileSystemTraverser = function(path, extension) {

  try {

    this.path = path;
    this.files = [];
    this.extension =  '.' + extension;
    this.traverse(this.path);

  } catch (e) {}
}
/**
 * FileSystemTraverser#traverse recursive method for traversing all objects in a path
 * @param {String} path - the filepath to traverse
 * */
FileSystemTraverser.prototype.traverse = function(path) {

  // get all the files in the directory
  var files = fs.readdirSync(path);

  for(let i = 0; i < files.length; ++i) {

    var file =  files[i];
    //we need this to traverse the next level
    var pathToFile = path + '/' + file;
    if (this.isDirectory(pathToFile)) {
      //if it's a directory than traverse it too
      this.traverse(pathToFile);
    } else {
      //it must be a file
      if (file.endsWith(this.extension)) {
        //it's the extension we are looking for so record it
        var object = file.replace(this.extension, '');
        this.files.push({filename:pathToFile, object: object });
      }
    }
  }
};

/**
 * FileSystemTraverser#isDirectory - a helper method for determining synchronously if a file is a directory or not
 * @param {String} path - the absolute path of that file/directory
 * */

FileSystemTraverser.prototype.isDirectory = function (path) {
  return fs.lstatSync(path).isDirectory()
};

/**
 * FileSystemTraverser#getFileHashList - a conveniece method for listing all the objects of a given extension in a
 * directory path. The default extension is javascript files
 * @param {String} path - the absolute path of that file/directory
 * */
FileSystemTraverser.getFileHashList = function(path) {
  var traverser = new FileSystemTraverser(path, 'js');
  return traverser.files;
};

Builder.FileSystemTraverser = FileSystemTraverser;

/**
 * getTrace - is a helper for stringifying stack traces encountered while building the intuition project
 * @param {Error} error - the error to get the stack trace of
 * @return {Array<String>} - the stack trace of the encourntered error as an array
 * */
var getTrace = function (error) {
  var stack = error.stack;
  var trace = [];
  for(var i = 0; i < stack.length; ++i) {
    trace.push(stack[i].getFileName() + ':' +stack[i].getLineNumber());
  }
  return trace;
}

module.exports = Builder;
