var co = require('co');

/**
 * Intuition is our super class it is simply a blank function with only a constructor property
 */
var Intuition  = function() {
  this.init = new Function;
};

/**
 * _super is never directly called but bound to with a prototype and depth
 * @param {Number} depth - is the level in the inheritance chain that super is bound to
 * @returns {Intuition} _super - is a set of methods belonging to a parent class bound to the current Intuition context
 */
Intuition.__super = function(depth) {
  var _super = {};

  // if the depth is a number greater than zero, then a _super class exists
  if(depth && depth > 0) {
    // the parent is like a linked list. When we finally reach the parent with depth, 1 less than the context depth
    // we know we have found the correct parent
  	var parent = this;

  	while(parent.__depth !== depth - 1) {

  		parent = parent._parent;
  	}
    // iterate through all the methods of the parent and attach them to the _super object with the context of the
    // the current instance's context
  	for(var method in parent) {
    	if(parent[method] && typeof parent[method] === 'function' && method !== 'super') {
        _super[method] = parent[method].bind(this);
      }
  	}
  }
  return _super;
};

/**
 * extend enables classical inheritance in javascript, anything that extends an Intuition Class can be Extended
 * @param {Function} constructor - is a function that retuns an object literal of inheritable methods
 * @returns {Intuition<Class>} Class - is a set of methods belonging to a parent class for the current Intuition context
 */
Intuition.extend = function(constructor) {

  if(!constructor) {
    throw new Error('Cannot extend a class without a primary constructor');
  }

  // we must create a class to describe what we are extending and to return
  var Class = function() {

   /** calls the primary constructor of the parent. Gererally arguments will be an empty array. In most cases only the
    * secondary constructor (init) will take method arguments
    */
  	Class.parent.apply(this, arguments);

    //creates an instance of the parent
  	this._parent = new Class.parent(arguments);

    // if this is the first object in the inheritance chain (i.e. it directly extends intuition) then the __depth
    // will be set to 0. After that the depth will be __depth of the parent incremented by 1
  	if(this.__depth === undefined && this._parent.__depth === undefined) {
    	this.__depth = 0;
    } else {
    	this.__depth = this._parent.__depth + 1;
    }

    // finally get an object literal of the "inheritable" methods. If the constructor doesn't return anything, then
    // create a blank object literal
    var instanceMethods = constructor.apply(this, arguments) || {};

    // if the secondary constructor is missing (init), then create a secondary constructor that calls the init method
    // of the parent
    if (!instanceMethods.init) {
      instanceMethods.init = function() {
        if(this.super.init) {
          return this.super.init.apply(this, arguments);
        }
      }
    }

   /**
    * redefineSuper is a simple helper method helping us redefine super on the fly. It should only be
    * invoked with .call
    * @param {Number} the current depth in the inheritance chain
    *  */
    var redefineSuper = function(depth) {
      Object.defineProperty(this, 'super', {
        get: function() {
          return Intuition.__super.call(this, depth);
        },
        configurable: true
      });
    }

    // this is a list of all the "inheritable" instance methods from the constructor function
    this.__instanceMethods = instanceMethods;

    //iterate through all parent methods
    for(var method in this._parent) {

      // we must capture both the method and depth in a closure since javascript is JIT compiled and method and depth
      // would not retain their values
      (function(method, depth){

        // if this is a method (not a property), we should copy the values into the current instance
        if(this._parent[method] && typeof this._parent[method] === 'function') {
          // Is this a special intuition method defined in the Intuition.__prototype ... it to this object
          // we won't be able to call super on these
          if (Intuition.__prototype[method]) {

            this[method] = function() {
              return Intuition.prototype[method].apply(this, arguments);
            }

          } else if(!instanceMethods[method]) {
            // if this level of inheritance does not implement the method, then try calling it in the parent class
            this[method] = function() {
              // we should dynamically redefine super every time a method is called so as to get the right depth
              redefineSuper.call(this, depth);
              return this.super[method].apply(this, arguments)
            }
          }
        }

      }).call(this, method, this.__depth);
    }

    // finally iterate through this level's methods so as to overwrite any parent methods
    for(var method in instanceMethods) {

      // once again capture the depth and method in a closure
      (function(method, depth){

        // does this class implement dynamically generated coroutines? And is this a generator?
        if (this.__enableCoroutines && this.__isAGeneratorFunction(method)) {

          (function(generatorMethod){
            // we now have captured the generatorMethod in a closure
            instanceMethods[method] = function() {
              // we can treat this like a regular coroutine method now
              // since this is a $coroutine, we do not need to worry about argument passing, since it will
              // automatically recieve [resolve, reject] as its arguments
              redefineSuper.call(this, depth);
              return this.$coroutine(generatorMethod);
            }
            // now add back the generated coroutine to the instanceMethods
            this.__instanceMethods[method] = instanceMethods[method];

          }).call(this,instanceMethods[method]);
        }

        this[method] = function() {
          // this is the simplest case, where we can just call our method
          redefineSuper.call(this, depth);
          return instanceMethods[method].apply(this, arguments);

        };

      }).call(this, method, this.__depth);
    }
    // finally create a reference to the Class itself in the instance
    this.__class = Class;

  };
  // the current context of this is referrring the object be extending (i.e. the parent)
  Class.parent = this;
  // this allows for instanceof to behave as expected
  Class.prototype = Object.create(Class.parent.prototype);
  // finally copy over all the the current classes static methods into that of the inheriting class
  Intuition.copyStaticMethods.call(Class, this);

  (function(Class) {
    // creates a way to instantiate instances of the class just created
    // new is a static method
    Class.new = function() {
      var object = new Class();
      object.__class__ = Class.__name__;
      //if a secondary constructor is present invoke it and return its value
      var returnValue = object.init.apply(object, arguments);
      if(returnValue) return returnValue;
      return object;
    };

  })(Class)

  return Class;

};

/**
 * copyStaticMethods copies the static methods of one level of inheritance into the inheriting class
 * @param {Intuition<Class>} _class
 */
Intuition.copyStaticMethods = function(_class) {

  for(var prop in _class) {
    if(!this[prop]) this[prop] = _class[prop];
  }

  if(_class.parent) Intuition.copyStaticMethods.call(_class, _class.parent);
};


/**
* getStaticCallerNameSpace returns the callerfile of a static method. This is only ever invoked in a compiled app
*/
Intuition.getStaticCallerNameSpace = function() {

  var caller = arguments.callee.caller;

  while (caller.length === 1) {
    caller = caller.caller;
  }

  var callerfile = caller.arguments[2].id;

  for(var i = 0; i < imports.builder.applicationList.length; ++i) {
    if(imports.builder.applicationList[i].filename == callerfile) break;
  }
  return imports.builder.applicationList[i].namespaceString;
};

/**
 * Intuition.__prototype is the pseudo-prototype of the class
 * */
Intuition.__prototype = {

  /**
   * __isAGeneratorFunction
   * @param {String} property on an object
   * @returns {Bool} whether a property is a generator
   * */
  __isAGeneratorFunction: function(property) {
    try {
      var method = this.__instanceMethods[property];
      return method.constructor.name === 'GeneratorFunction';
    } catch (e) {}
    return false;
  },

  /**
   * stub allows a user to mock the values of return and resolves during testing for any instance of an Intuition object
   * @param {String} methodName - a method to override
   * @returns {Intuition} returns a modified version of itself
   * */
  stub: function(methodName) {
    (function(methodName){
      //creates a dynamic method stubbing the return value if it's a synchronous method
      this['returns'] = function(stubbedValue) {
         this[methodName] = function() {
           return stubbedValue;
         }
      }
      //creates a dynamic method stubbing the resolve value if it's a promise
      this['resolves'] = function(stubbedValue) {
         this[methodName] = function() {
           return new Promise(function(resolve) {
              resolve(stubbedValue);
           });
         }
      }
    }).call(this, methodName);
    return this;
  },

  /**
   * $coroutine a method that wraps coroutines in promise and allows for continuous chaining of coroutines
   * @param {Function*} generator - a function that will be run as a coroutine
   * @returns {Promise} returns a modified version of itself
   * */
  $coroutine: function(generator) {
     return new Promise(function(resolve, reject){
        var generatorMethod = generator.bind(this, resolve, reject);
        co(generatorMethod)
        .catch(function(error) {
           reject(error);
        }.bind(this));
     }.bind(this));
  }
}

Intuition.prototype = Intuition.__prototype;

//adds $coroutines to an Intuition object's static methods
Intuition.$coroutine = function(generator) {
   return Intuition.prototype.$coroutine.call(this, generator);
}

module.exports = Intuition;